/* $Id: ike_pfkey.c,v 1.63 2006/07/05 06:18:11 sinoue Exp $ */

/*
 * Copyright (C) 2004 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * IKE-PFKEY interface bridge
 */

#include <config.h>

#include <stdio.h>
#include <string.h>
#ifdef HAVE_INTTYPES_H
#  include <inttypes.h>
#endif
#include <sys/types.h>
#if TIME_WITH_SYS_TIME
#  include <sys/time.h>
#  include <time.h>
#else
#  if HAVE_SYS_TIME_H
#    include <sys/time.h>
#  else
#    include <time.h>
#  endif
#endif
#include <arpa/inet.h>

#include "racoon.h"
#include "ike_pfkey.h"
#include "isakmp_impl.h"
#include "ikev2_impl.h"
#include "debug.h"

extern int debug_pfkey;
static void dump_param(char *, struct rcpfk_msg *);

static int sadb_getspi(struct rcpfk_msg *);
static int sadb_acquire_error(struct rcpfk_msg *);
static int sadb_update(struct rcpfk_msg *);
static int sadb_add(struct rcpfk_msg *);
static int sadb_responder_error(struct rcpfk_msg *);
static int sadb_delete(struct rcpfk_msg *);

static int
null_proc()
{
	return 0;
}

/* sadb_initiator_request_method used in response to SADB_ACQUIRE */
struct sadb_request_method sadb_initiator_request_method = {
	sadb_getspi,
	sadb_acquire_error,
	sadb_update,
	sadb_add,
	sadb_delete,
};

/* sadb_responder_request_method for use when receiving IKE_SA_INIT packet */
struct sadb_request_method sadb_responder_request_method = {
	sadb_getspi,
	sadb_responder_error,
	sadb_update,
	sadb_add,
	sadb_delete,
};

/* sadb_rekey_request_method for use when rekeying soft-expired IPsec SA */
struct sadb_request_method sadb_rekey_request_method = {
	sadb_getspi,
	sadb_responder_error,
	sadb_update,
	sadb_add,
	sadb_delete,
};

/* sadb_null_method for informational exchange SA */
struct sadb_request_method sadb_null_method = {
	null_proc, null_proc, null_proc, null_proc, null_proc
};

static SADB_LIST_HEAD(sadb_request_list_head, sadb_request) sadb_request_list_head;

static int pfkey_socket;
static u_int32_t sadb_msg_seq;

static int sadb_getspi_callback(struct rcpfk_msg *param);
static int sadb_expire_callback(struct rcpfk_msg *param);
static int sadb_acquire_callback(struct rcpfk_msg *param);
static int sadb_delete_callback(struct rcpfk_msg *param);
#ifdef SADB_X_MIGRATE
static int sadb_x_migrate_callback(struct rcpfk_msg *param);
#endif

static struct rcpfk_cb ike_rcpfk_callback = {
	sadb_getspi_callback,
	0,		/* sadb_update_callback, */
	0,		/* sadb_add_callback, */
	sadb_expire_callback,
	sadb_acquire_callback,
	sadb_delete_callback,
	0,		/* sadb_get_callback, */
	0,		/* sadb_spdupdate_callback, */
	0,		/* sadb_spdadd_callback, */
	0,		/* sadb_spddelete_callback, */
	0,		/* sadb_spddelete2_callback, */
	0,		/* sadb_spdexpire_callbcak, */
	0,		/* sadb_spdget_callback */
	0,		/* sadb_spddump_callback */
#ifdef SADB_X_MIGRATE
	sadb_x_migrate_callback,
#endif
};

int
sadb_init(void)
{
	struct rcpfk_msg param;

	SADB_LIST_INIT(&sadb_request_list_head);
	if (debug_pfkey)
		return 0;

	param.flags = 0;

	if (rcpfk_init(&param, &ike_rcpfk_callback) != 0)
		return -1;
	pfkey_socket = param.so;
	TRACE((PLOGLOC, "pfkey_socket: %d\n", pfkey_socket));
	return 0;
}

#ifdef DEBUG
void
sadb_list_dump(void)
{
	struct sadb_request *req;

	plog(PLOG_DEBUG, PLOGLOC, 0, "sadb request list:\n");
	for (req = SADB_LIST_FIRST(&sadb_request_list_head);
	     !SADB_LIST_END(req);
	     req = SADB_LIST_NEXT(req)) {
		plog(PLOG_DEBUG, PLOGLOC, 0,
		     "req %p method:%p seqno:%lx getspi_response:%p expired:%p sa:%p\n",
		     req, req->method, (unsigned long)req->seqno,
		     req->getspi_response, req->expired, req->sa);
	}
	plog(PLOG_DEBUG, PLOGLOC, 0, "end\n");
}
#endif

int
sadb_socket(void)
{
	return pfkey_socket;
}

u_int32_t
sadb_new_seq(void)
{
	return ++sadb_msg_seq;
}

static void
log_rcpfk_error(char *msg, struct rcpfk_msg *param)
{
	if (param->eno) {
		isakmp_log(0, 0, 0, 0,
			   PLOG_INTERR, PLOGLOC,
			   "%s: %s\n", msg, param->estr);
	} else {
		isakmp_log(0, 0, 0, 0,
			   PLOG_INTERR, PLOGLOC,
			   "%s: unknown error\n", msg);
	}
}

void
sadb_poll(void)
{
	struct rcpfk_msg rcpfk_param;

	rcpfk_param.so = pfkey_socket;
	if (rcpfk_handler(&rcpfk_param) != 0) {
		log_rcpfk_error("sadb_poll", &rcpfk_param);
	}
}

void
sadb_request_initialize(struct sadb_request *req, struct sadb_request_method *m,
			u_int32_t seqno, int (*getspi_response)(),
			void (*expired)(), void *sa)
{
	req->method = m;
	req->seqno = seqno;
	req->getspi_response = getspi_response;
	req->expired = expired;
	req->sa = sa;
	SADB_LIST_LINK(&sadb_request_list_head, req);
}

void
sadb_request_finish(struct sadb_request *req)
{
	TRACE((PLOGLOC, "%p\n", req));
	if (req->link.tqe_prev != 0)	/* initialized? */
		SADB_LIST_REMOVE(&sadb_request_list_head, req);
}

/*
 * Send a SADB_GETSPI message
 */
static int
sadb_getspi(struct rcpfk_msg *param)
{
	int err;

	TRACE((PLOGLOC, "sadb_getspi: seq=0x%x, satype=%d\n",
	       param->seq, param->satype));

	param->so = pfkey_socket;
	param->eno = 0;
	err = rcpfk_send_getspi(param);
	if (err)
		log_rcpfk_error("sadb_getspi", param);
	return err;
}

/*
 * send SADB_ACQUIRE with error to inform kernel of SA creation failure
 */
static int
sadb_acquire_error(struct rcpfk_msg *param)
{
	int err;

	TRACE((PLOGLOC,
	       "sadb_acquire_error: seq=0x%x, satype=%d, spi=0x%x, errno=%d\n",
	       param->seq, param->satype, ntohl(param->spi), param->eno));

	/* param: so, satype, seq, eno */
	param->so = pfkey_socket;
	err = rcpfk_send_acquire(param);
	if (err)
		log_rcpfk_error("sadb_acquire_error", param);
	return err;
}

static void
sadb_log_add(char *op, struct rcpfk_msg *param)
{
	if (param->satype == RCT_SATYPE_ESP) {
		INFO((PLOGLOC,
		      "%s src=%s dst=%s satype=%s samode=%s spi=0x%x authtype=%s enctype=%s lifetime soft time=%"
		      PRIu64 " bytes=%" PRIu64 " hard time=%" PRIu64 " bytes=%" PRIu64 "\n",
		      op, rcs_sa2str(param->sa_src),
		      rcs_sa2str(param->sa_dst), rct2str(param->satype),
		      rct2str(param->samode), ntohl(param->spi),
		      rct2str(param->authtype), rct2str(param->enctype),
		      param->lft_soft_time, param->lft_soft_bytes,
		      param->lft_hard_time, param->lft_hard_bytes));
	} else {
		INFO((PLOGLOC,
		      "%s src=%s dst=%s satype=%s samode=%s spi=0x%x authtype=%s lifetime soft time=%"
		      PRIu64 " bytes=%" PRIu64 " hard time=%" PRIu64 " bytes=%" PRIu64 "\n",
		      op, rcs_sa2str(param->sa_src),
		      rcs_sa2str(param->sa_dst), rct2str(param->satype),
		      rct2str(param->samode), ntohl(param->spi),
		      rct2str(param->authtype), param->lft_soft_time,
		      param->lft_soft_bytes, param->lft_hard_time,
		      param->lft_hard_bytes));
	}
}

/* send SADB_UPDATE */
static int
sadb_update(struct rcpfk_msg *param)
{
	int err;

	sadb_log_add("SADB_UPDATE", param);
	IF_TRACE(dump_param("sadb_update", param));

	/* param:
	 * so, satype, seq, spi, wsize, authtype, [enctype,] saflags, samode, reqid,
	 * lft_hard_time, lft_hard_bytes, lft_soft_time, lft_soft_bytes,
	 * sa_src, pref_src, sa_dst, pref_dst, ul_proto,
	 * [enckey, enckeylen], authkey, authkeylen
	 */
	param->so = pfkey_socket;
	param->eno = 0;
	err = rcpfk_send_update(param);
	if (err)
		log_rcpfk_error("sadb_update", param);
	return err;
}

/* send SADB_ADD */
static int
sadb_add(struct rcpfk_msg *param)
{
	int err;

	sadb_log_add("SADB_ADD", param);
	IF_TRACE(dump_param("sadb_add", param));

	param->so = pfkey_socket;
	param->eno = 0;
	err = rcpfk_send_add(param);
	if (err)
		log_rcpfk_error("sadb_add", param);
	return err;
}

/*
 * acquire_error for responder
 */
static int
sadb_responder_error(struct rcpfk_msg *param)
{
	/* just ignore since there's no corresponding SADB_ACQUIRE */
	TRACE((PLOGLOC,
	       "sadb_responder_error: seq=0x%x, satype=%d, spi=0x%x, errno=%d\n",
	       param->seq, param->satype, ntohl(param->spi), param->eno));
	return 0;
}

/*
 * send SADB_DELETE
 */
static int
sadb_delete(struct rcpfk_msg *rc)
{
	int err;

	INFO((PLOGLOC,
	      "SADB_DELETE src=%s dst=%s satype=%s spi=0x%08x\n",
	      rcs_sa2str(rc->sa_src), rcs_sa2str(rc->sa_dst), rct2str(rc->satype),
	      ntohl(rc->spi)));
	TRACE((PLOGLOC,
	       "sadb_delete: sa_src=%s, sa_dst=%s, satype=%d (%s), spi=0x%x\n",
	       rcs_sa2str(rc->sa_src), rcs_sa2str(rc->sa_dst), rc->satype,
	       rct2str(rc->satype), ntohl(rc->spi)));

	/* param: so, satype, spi, sa_src, sa_dst */
	/* XXX
	 *        pref_dst,pref_src must be <= addrlen, eventhough the values aren't used
	 */
	rc->so = pfkey_socket;
	rc->eno = 0;
	rc->seq = 0;		/* ??? it seems ignored */
	rc->pref_src = rc->pref_dst = 0;	/* ??? */
	err = rcpfk_send_delete(rc);
	if (err)
		log_rcpfk_error("sadb_delete", rc);
	return err;
}

/*
 * receive SADB_GETSPI message from kernel
 */
static int
sadb_getspi_callback(struct rcpfk_msg *param)
{
	/* param: seq, satype, spi, sa_src, sa_dst */

	struct sadb_request *req;
	u_int32_t seq;

	TRACE((PLOGLOC,
	       "sadb_getspi_callback: seq=%d, spi=0x%x, satype=%d, sa_src=%s, sa_dst=%s\n",
	       param->seq, ntohl(param->spi), param->satype,
	       rcs_sa2str(param->sa_src), rcs_sa2str(param->sa_dst)));

	/* find sadb_request by param->seq */
	seq = param->seq;
	for (req = SADB_LIST_FIRST(&sadb_request_list_head);
	     !SADB_LIST_END(req);
	     req = SADB_LIST_NEXT(req)) {
		if (req->seqno == seq) {
			/* call request->getspi_response(param) */
			return (req->getspi_response ?
				(*req->getspi_response)(req, param->sa_src,
						        param->sa_dst,
							(u_int)param->satype,
							ntohl(param->spi)) :
				-1);
		}
	}

	/* couldn't find corresponding SA */
	isakmp_log(0, 0, 0, 0, PLOG_INTERR, PLOGLOC,
		   "PF_KEY SADB_GETSPI message does not have corresponding request. (ignored)\n");
	return 0;
}

#if 0
/* not used */
/* called when other KMd issued SADB_UPDATE */
static int
sadb_update_callback(struct rcpfk_msg *param)
{
	/* param: seq, satype, spi, sa_src, sa_dst, samode */
	/* lifetime??? address(P)??? identity??? */

	return 0;
}
#endif

#if 0
/* not used */
/* called when other KMd issued SADB_UDPATE */
static int
sadb_add_callback(struct rcpfk_msg *param)
{
	/* param: seq, satype, spi, sa_src, sa_dst, samode */
	/* lifetime??? identity?? sensitivity?? */

	return 0;
}
#endif

/*
 * called when kernel SA expires
 */
static int
sadb_expire_callback(struct rcpfk_msg *param)
{
	/* param: seq, satype, spi, sa_src, sa_dst, samode, expired(hard?2:1) */
	/* lifetime(C)??? */

	struct sadb_request *req;
	int satype;
	struct ikev2_child_sa *child_sa;
	struct rcf_policy *policy;
	struct sockaddr *localaddr;
	struct sockaddr *remoteaddr;
	struct prop_pair *proposal;

	plog(PLOG_INFO, PLOGLOC, 0,
	     "received PFKEY_EXPIRE seq=%d sa_dst=%s spi=0x%x satype=%s samode=%s expired=%d\n",
	     param->seq, rcs_sa2str(param->sa_dst), ntohl(param->spi),
	     rct2str(param->satype), rct2str(param->samode), param->expired);

	/* #ifdef __linux__ ??? */
	/* Linux/USAGI generates soft-expire regardless it was used or not */
	TRACE((PLOGLOC, "allocated: %ld\n", (long)param->lft_current_alloc));
	if (param->expired == 1 && param->lft_current_alloc == 0) {
		TRACE((PLOGLOC, "ignoring soft expire\n"));
		return 0;
	}
	/* #endif */

	switch (param->satype) {
	case RCT_SATYPE_ESP:
		satype = IKEV2PROPOSAL_ESP;
		break;
	case RCT_SATYPE_AH:
		satype = IKEV2PROPOSAL_AH;
		break;
	case RCT_SATYPE_IPCOMP:
		goto done;
	default:
		isakmp_log(0, 0, 0, 0,
			   PLOG_INTERR, PLOGLOC,
			   "unknown satype %d (%s) in sadb_expire callback parameter\n",
			   param->satype, rct2str(param->satype));
		goto done;
	}

	/* start rekeying */
	/* find sadb_request by spi, sa_dst */
	for (req = SADB_LIST_FIRST(&sadb_request_list_head);
	     !SADB_LIST_END(req); req = SADB_LIST_NEXT(req)) {
		child_sa = (struct ikev2_child_sa *)req->sa;
		if (child_sa->state != IKEV2_CHILD_STATE_MATURE) {
			TRACE((PLOGLOC, "child_sa %p state %d skipped\n",
			       child_sa, child_sa->state));
			continue;
		}
		policy = child_sa->selector->pl;
		if (policy->my_sa_ipaddr) {
			if (policy->my_sa_ipaddr->type != RCT_ADDR_INET) {
				TRACE((PLOGLOC, "unexpected type\n"));
				continue;
			}
			localaddr = policy->my_sa_ipaddr->a.ipaddr;
		} else {
			localaddr = child_sa->parent->local;
		}
		if (policy->peers_sa_ipaddr &&
		    !rcs_is_addr_rw(policy->peers_sa_ipaddr)) {
			if (policy->peers_sa_ipaddr->type != RCT_ADDR_INET) {
				TRACE((PLOGLOC, "unexpected type\n"));
				continue;
			}
			remoteaddr = policy->peers_sa_ipaddr->a.ipaddr;
		} else {
			remoteaddr = child_sa->parent->remote;
		}

		if (rcs_cmpsa_wop(localaddr, param->sa_dst) == 0) {
			for (proposal = child_sa->my_proposal[1];
			     proposal;
			     proposal = proposal->next) {
				struct isakmp_pl_p	*prop;

				prop = proposal->prop;
				if (prop->proto_id == satype
				    && *(u_int32_t *)(prop + 1) == param->spi) {
					if (req->expired) {
						TRACE((PLOGLOC,
						       "expiring child_sa %p\n",
						       child_sa));
						(*req->expired)(req,
								(u_int)param->satype,
								ntohl(param->spi),
								param->sa_src,
								param->sa_dst,
								param->samode,
								param->expired);
					} else {
						TRACE((PLOGLOC,
						       "child_sa %p req %p does not have expire method\n",
						       child_sa, req));
					}
					goto done;
				}
			}
		} else if (rcs_cmpsa_wop(remoteaddr, param->sa_dst) == 0) {
			TRACE((PLOGLOC,
			       "expire message was for outbound ipsec_sa of child_sa %p\n",
			       child_sa));
			goto done;
		}
	}
	/* couldn't find corresponding SA */
	isakmp_log(0, 0, 0, 0, PLOG_INTWARN, PLOGLOC,
		   "PF_KEY SADB_EXPIRE message does not have corresponding request. (ignored)\n");

      done:
	return 0;
}

/*
 * called when the kernel generates SADB_ACQUIRE message
 */
static int
sadb_acquire_callback(struct rcpfk_msg *param)
{
	/* param: seq, satype, sa_src, sa_dst, samode, selid */
	/* address(P)??? pid?? identity??? proposal??? */

	TRACE((PLOGLOC,
	       "sadb_acquire_callback: seq=%d satype=%d sa_src=%s sa_dst=%s samode=%d selid=%d\n",
	       param->seq, param->satype, rcs_sa2str(param->sa_src),
	       rcs_sa2str(param->sa_dst), param->samode, param->slid));

	param->eno = isakmp_initiate(&sadb_initiator_request_method,
				     param->slid,
				     param->seq, param->satype,
				     param->sa_src, param->sa_dst,
				     param->sa2_src);
	if (param->eno != 0)
		sadb_acquire_error(param);
	return 0;
}

/*
 * called when the kernel generates SADB_DELETE message
 */
static int
sadb_delete_callback(struct rcpfk_msg *param)
{
	/* param: seq, satype, spi, sa_src, sa_dst, samode */

	/* similar to expire ? */

	plog(PLOG_INFO, PLOGLOC, 0,
	     "received PFKEY_DELETE seq=%d satype=%s spi=0x%x\n",
	     param->seq, rct2str(param->satype), ntohl(param->spi));
	return 0;
}

#if 0
/* not used */
/* called when other KMd issued SADB_X_SPDUPDATE */
static int
sadb_spdupdate_callback(struct rcpfk_msg *param)
{
	/* param: selid */

	return 0;
}
#endif

#if 0
/* not used */
/* called when other KMd issued SADB_X_SPDADD */
static int
sadb_spdadd_callback(struct rcpfk_msg *param)
{
	/* param: selid */
	return 0;
}
#endif

#if 0
/* called when other KMd issued SADB_X_SPDDELETE */
static int
sadb_spddelete_callback(struct rcpfk_msg *param)
{
	/* param: selid */
	return 0;
}
#endif

#if 0
/* called when kernel SP expires */
static int
sadb_spdexpire_callback(struct rcpfk_msg *param)
{
	/* param: selid */
	/* address(SD)? lifetime(CH)? */

	return 0;
}
#endif

#ifdef SADB_X_MIGRATE
#include <netinet/in.h>
/* called when kernel issued SADB_X_MIGRATE */
static int
sadb_x_migrate_callback(struct rcpfk_msg *param)
{
	
	struct ikev2_sa *ike_sa;
	struct ikev2_child_sa *child_sa;

	struct sockaddr *local, *remote;
	struct rcf_policy *policy;
	char *a, *b;
	

	a = strdup(rcs_sa2str(param->sa2_src));
	b = strdup(rcs_sa2str(param->sa2_dst));

	TRACE((PLOGLOC,
		"sadb_x_migrate_callback: dir=%s, sa_src=%s, sa_dst=%s, sa2_src=%s, sa2_dst=%s\n",
	 	rct2str(param->dir),	
		rcs_sa2str(param->sa_src), rcs_sa2str(param->sa_dst),
		rcs_sa2str(param->sa2_src), rcs_sa2str(param->sa2_dst)));

	racoon_free(a);
	racoon_free(b);

	local = (struct sockaddr *)param->sa_src;
	remote = (struct sockaddr *)param->sa_dst;
	if ((rcs_cmpsa(local, param->sa2_src) == 0) &&
	     (rcs_cmpsa(remote, param->sa2_dst) == 0))
	  return 0;

	if (param->dir != RCT_DIR_OUTBOUND)
		return 0;

	ike_sa = ikev2_find_sa_by_addr(remote);
	if (ike_sa == NULL)
	  return 0;

	plog(PLOG_INFO, PLOGLOC, 0,
		"move ikev2_sa(%p): from %s -> %s\n",
		ike_sa, rcs_sa2str(local), rcs_sa2str(remote));
	  
	local = (struct sockaddr *)param->sa2_src;
	remote = (struct sockaddr *)param->sa2_dst;

	switch (ike_sa->remote->sa_family) {
	case AF_INET:
		((struct sockaddr_in *)ike_sa->local)->sin_addr =
			((struct sockaddr_in *)local)->sin_addr;
		((struct sockaddr_in *)ike_sa->remote)->sin_addr =
			((struct sockaddr_in *)remote)->sin_addr;
#ifdef INET6
	case AF_INET6:
		memcpy(&((struct sockaddr_in6 *)ike_sa->local)->sin6_addr, 
			&((struct sockaddr_in6 *)local)->sin6_addr,
			sizeof(struct in6_addr));
		memcpy(&((struct sockaddr_in6 *)ike_sa->remote)->sin6_addr,
			&((struct sockaddr_in6 *)remote)->sin6_addr,
			sizeof(struct in6_addr));
		break;
#endif
	default:
		return -1;
	}
	plog(PLOG_INFO, PLOGLOC, 0,
		"move ikev2_sa(%p): to %s -> %s\n",
		ike_sa, rcs_sa2str(local), rcs_sa2str(remote));

	
	for (child_sa = IKEV2_CHILD_LIST_FIRST(&ike_sa->children);
	     !IKEV2_CHILD_LIST_END(child_sa);
	     child_sa = IKEV2_CHILD_LIST_NEXT(child_sa)){

	  if (child_sa->selector) 
	    if (param->reqid == child_sa->selector->reqid) {
	      
	      switch (ike_sa->remote->sa_family) {
	      case AF_INET:
		if (child_sa->local)
		  ((struct sockaddr_in *)child_sa->local)->sin_addr =
		    ((struct sockaddr_in *)ike_sa->local)->sin_addr;
		if (child_sa->remote)
		  ((struct sockaddr_in *)child_sa->remote)->sin_addr =
		    ((struct sockaddr_in *)ike_sa->remote)->sin_addr;
		
		policy = child_sa->selector->pl;
		
		if (policy->my_sa_ipaddr) {
		  if (policy->my_sa_ipaddr->type != RCT_ADDR_INET) {
		    TRACE((PLOGLOC, "unexpected type\n"));
		    continue;
		  }
		  ((struct sockaddr_in *)policy->my_sa_ipaddr->a.ipaddr)->sin_addr =
		    ((struct sockaddr_in *)ike_sa->local)->sin_addr;
		}
		
		if (policy->peers_sa_ipaddr) {
		  if (policy->peers_sa_ipaddr->type != RCT_ADDR_INET) {
		    TRACE((PLOGLOC, "unexpected type\n"));
		    continue;
		  }
		  ((struct sockaddr_in *)policy->peers_sa_ipaddr->a.ipaddr)->sin_addr =
		    ((struct sockaddr_in *)ike_sa->local)->sin_addr;
		}
		break;
#ifdef INET6
	      case AF_INET6:
		if(child_sa->local)
		  memcpy(&((struct sockaddr_in6 *)child_sa->local)->sin6_addr,
			 &((struct sockaddr_in6 *)ike_sa->local)->sin6_addr,
			 sizeof(struct in6_addr));
		if(child_sa->remote)
		  memcpy(&((struct sockaddr_in6 *)child_sa->remote)->sin6_addr,
			 &((struct sockaddr_in6 *)ike_sa->remote)->sin6_addr,
			 sizeof(struct in6_addr));
		
		policy = child_sa->selector->pl;
		
		if (policy->my_sa_ipaddr) {
		  if (policy->my_sa_ipaddr->type != RCT_ADDR_INET) {
		    TRACE((PLOGLOC, "unexpected type\n"));
		    continue;
		  }
		  memcpy(&((struct sockaddr_in6 *)policy->my_sa_ipaddr->a.ipaddr)->sin6_addr, 
			 &((struct sockaddr_in6 *)ike_sa->local)->sin6_addr,
			 sizeof(struct in6_addr));
		}
		
		if (policy->peers_sa_ipaddr) {
		  if (policy->peers_sa_ipaddr->type != RCT_ADDR_INET){
		    TRACE((PLOGLOC, "unexpected type\n"));
		    continue;
		  }
		  memcpy(&((struct sockaddr_in6 *)policy->peers_sa_ipaddr->a.ipaddr)->sin6_addr,
			 &((struct sockaddr_in6 *)ike_sa->remote)->sin6_addr,
			 sizeof(struct in6_addr));
		}
		break;
#endif
	      }
	      plog(PLOG_INFO, PLOGLOC, 0,
		   "move child_sa(%p)\n", child_sa);
	    }	
	}
	return 0;
}
#endif

/* #ifdef DEBUG */
static int sadb_debug_getspi(struct rcpfk_msg *param);
static int sadb_debug_acquire_error(struct rcpfk_msg *param);
static int sadb_debug_update(struct rcpfk_msg *param);
static int sadb_debug_add(struct rcpfk_msg *param);
static int sadb_debug_delete(struct rcpfk_msg *param);

struct sadb_request_method sadb_debug_method = {
	sadb_debug_getspi,
	sadb_debug_acquire_error,
	sadb_debug_update,
	sadb_debug_add,
	sadb_debug_delete,
};

u_int32_t debug_spi = 0x10000;

static int
sadb_debug_getspi(struct rcpfk_msg *param)
{
	int err;

	TRACE((PLOGLOC, "sadb_debug_getspi: seq=0x%x, satype=%d\n",
	       param->seq, param->satype));

	param->spi = htonl(debug_spi++);
	err = sadb_getspi_callback(param);
	TRACE((PLOGLOC, "sadb_getspi_callback retval %d\n", err));
	return 0;
}

static int
sadb_debug_acquire_error(struct rcpfk_msg *param)
{
	TRACE((PLOGLOC,
	       "sadb_debug_acquire_error: seq=0x%x, satype=%d, spi=0x%x, errno=%d\n",
	       param->seq, param->satype, ntohl(param->spi), param->eno));
	return 0;
}

static int
sadb_debug_update(struct rcpfk_msg *param)
{
	dump_param("sadb_debug_update", param);
	return 0;
}

static int
sadb_debug_add(struct rcpfk_msg *param)
{
	dump_param("sadb_debug_add", param);
	return 0;
}

static int
sadb_debug_delete(struct rcpfk_msg *param)
{
	dump_param("sadb_debug_delete", param);
	return 0;
}

/*
 * dump add/update parameters
 */
static void
dump_param(char *msg, struct rcpfk_msg *param)
{
	int i;
	char buf[BUFSIZ];
	char *bufp;
	ssize_t buflen;

#define	DUMP(x_)	do {						\
			    buflen -= strlen(bufp);			\
			    bufp += strlen(bufp);			\
			    if (buflen > 0) {				\
				x_;					\
			    }						\
			} while (0)

	buf[0] = '\0';
	bufp = &buf[0];
	buflen = sizeof(buf) - 1;
	DUMP(snprintf(bufp, buflen,
		      "%s: seq=0x%x, sa_src=%s/%d, sa_dst=%s/%d, "
		      "satype=%d (%s), spi=0x%x, wsize=%d, "
		      "authtype=%d (%s), enctype=%d (%s), saflags=0x%x, "
		      "samode=%d (%s), reqid=%d, "
		      "lifetime hard time %" PRIu64 ", bytes %" PRIu64 ", "
		      "lifetime soft time %" PRIu64 ", bytes %" PRIu64 ", "
		      "enckey len=%lu [",
		      msg,
		      param->seq,
		      rcs_sa2str(param->sa_src), param->pref_src,
		      rcs_sa2str(param->sa_dst), param->pref_dst,
		      param->satype, rct2str(param->satype),
		      ntohl(param->spi), param->wsize,
		      param->authtype, rct2str(param->authtype),
		      param->enctype, rct2str(param->enctype),
		      param->saflags,
		      param->samode, rct2str(param->samode),
		      param->reqid,
		      param->lft_hard_time, param->lft_hard_bytes,
		      param->lft_soft_time, param->lft_soft_bytes,
		      (unsigned long)param->enckeylen));
	for (i = 0; i < (int)param->enckeylen; ++i) {
		DUMP(snprintf(bufp, buflen,
			      "%02x", ((u_int8_t *)param->enckey)[i]));
	}
	DUMP(snprintf(bufp, buflen, "], authkey len=%lu [", 
		      (unsigned long)param->authkeylen));
	for (i = 0; i < (int)param->authkeylen; ++i) {
		DUMP(snprintf(bufp, buflen,
			      "%02x", ((u_int8_t *)param->authkey)[i]));
	}
	DUMP(snprintf(bufp, buflen, "]\n"));

	TRACE((PLOGLOC, "%s", buf));
}

#ifdef DEBUG
#include <sys/socket.h>
#include <netdb.h>

void
debug_initiate(char *addr, const char *selector_index)
{
	struct isakmp_acquire_request *req;
	struct addrinfo *res;
	int err;

	req = racoon_calloc(1, sizeof(*req));

	err = getaddrinfo(addr, 0, 0, &res);
	if (err) {
		fprintf(stderr, "getaddrinfo: %s\n", gai_strerror(err));
		return;
	}
	if (!res) {
		fprintf(stderr, "res is null\n");
		return;
	}
	if (!res->ai_addr) {
		fprintf(stderr, "res->ai_addr is null\n");
		return;
	}

	req->callback_method = &sadb_debug_method;
	req->request_msg_seq = 1;
	req->dst = rcs_sadup(res->ai_addr);
	(void)isakmp_initiate_cont(req, selector_index);

	freeaddrinfo(res);
}
#endif
/* #endif */
