/* $Id: ikev2_child.c,v 1.75 2006/07/07 12:34:45 fukumoto Exp $ */

/*
 * Copyright (C) 2004-2005 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <config.h>

#include <assert.h>
#include <sys/types.h>
#if TIME_WITH_SYS_TIME
#  include <sys/time.h>
#  include <time.h>
#else
#  if HAVE_SYS_TIME_H
#    include <sys/time.h>
#  else
#    include <time.h>
#  endif
#endif
#include <sys/errno.h>
#include <netinet/in.h>

#ifdef HAVE_INTTYPES_H
#  include <inttypes.h>
#endif

#include "racoon.h"

#include "isakmp.h"
#include "dhgroup.h"
#include "oakley.h"
#include "ikev2.h"
#include "isakmp_impl.h"
#include "ikev2_impl.h"
#include "ike_conf.h"
#include "crypto_impl.h"

#include "sockmisc.h"
#include "debug.h"

static void ikev2_child_getspi_done(struct ikev2_child_sa *);
static void ikev2_create_child_responder_cont(struct ikev2_child_sa *);

static int ikev2_add_ipsec_sa(struct ikev2_child_sa *,
			      struct ikev2_child_param *, struct prop_pair *,
			      struct prop_pair *);
static void ikev2_child_expire_callback(void *);
static void ikev2_expire_child(struct ikev2_child_sa *);

static vchar_t *compute_keymat(struct ikev2_sa *, vchar_t *, size_t, vchar_t *,
			       vchar_t *);

struct isakmp_domain ikev2_createchild_doi = {
	ikev2_check_spi_size,	/* check_spi_size */
	sizeof(isakmp_cookie_t),	/* ike_spi_size */
	FALSE,			/* check_reserved_fields */
	FALSE,			/* transform_number */
	ikev2_get_transforms,	/* get_transforms */
	ikev2_compare_transforms,
	ikev2_match_transforms
};

struct child_dispatch {
	void (*enter_state)(struct ikev2_child_sa *);
	enum ikev2_child_state next_state;
	void (*error)(struct ikev2_child_sa *, int);
};

static struct child_dispatch ikev2_child_dispatch[2][IKEV2_CHILD_STATE_NUM] = {
	{			/* initiator */
	 {0, IKEV2_CHILD_STATE_GETSPI, 0},
	 {0, IKEV2_CHILD_STATE_GETSPI_DONE, 0},
	 {ikev2_child_getspi_done, IKEV2_CHILD_STATE_WAIT_RESPONSE, 0},
	 {0, IKEV2_CHILD_STATE_MATURE, 0},
	 {0, IKEV2_CHILD_STATE_MATURE, 0},
	 {0, IKEV2_CHILD_STATE_EXPIRED, 0},
	 {0, IKEV2_CHILD_STATE_REQUEST_SENT, 0},
	 {0, IKEV2_CHILD_STATE_INVALID, 0},
	 },
	{			/* responder */
	 {0, IKEV2_CHILD_STATE_GETSPI, 0},
	 {0, IKEV2_CHILD_STATE_MATURE, 0},
	 {0, IKEV2_CHILD_STATE_INVALID, 0},
	 {0, IKEV2_CHILD_STATE_INVALID, 0},
	 {ikev2_create_child_responder_cont, IKEV2_CHILD_STATE_MATURE, 0},
	 {0, IKEV2_CHILD_STATE_EXPIRED, 0},
	 {0, IKEV2_CHILD_STATE_INVALID, 0},
	 {0, IKEV2_CHILD_STATE_INVALID, 0},
	 }
};

void
ikev2_child_state_set(struct ikev2_child_sa *child_sa,
		      enum ikev2_child_state state)
{
	struct child_dispatch *state_info;

	TRACE((PLOGLOC, "child_sa %p state %s -> %s\n",
	       child_sa,
	       ikev2_child_state_str(child_sa->state),
	       ikev2_child_state_str(state)));

	child_sa->state = state;
	if (state == IKEV2_CHILD_STATE_EXPIRED) {
		if (child_sa->timer)
			SCHED_KILL(child_sa->timer);
	}
	state_info =
		&ikev2_child_dispatch[child_sa->is_initiator ? 0 : 1][state];
	if (state_info->enter_state) {
		state_info->enter_state(child_sa);
	}
}

void
ikev2_child_state_next(struct ikev2_child_sa *child_sa)
{
	enum ikev2_child_state next_state;
	struct child_dispatch *state_info;

	assert(child_sa->state < IKEV2_CHILD_STATE_NUM);

	state_info =
		&ikev2_child_dispatch[child_sa->is_initiator ? 0 : 1][child_sa->state];
	next_state = state_info->next_state;

	if (next_state == IKEV2_CHILD_STATE_INVALID) {
		plog(PLOG_INTERR, PLOGLOC, 0,
		     "no next state for %s state %d\n",
		     (child_sa->is_initiator ? "initiator" : "responder"),
		     child_sa->state);
		next_state = IKEV2_CHILD_STATE_EXPIRED;	/* ??? */
	}

	ikev2_child_state_set(child_sa, next_state);
}

struct ikev2_child_sa *
ikev2_create_child_sa(struct ikev2_sa *ike_sa)
{
	struct ikev2_child_sa *child_sa;
	static u_int child_id;

	child_sa = racoon_calloc(1, sizeof(struct ikev2_child_sa));
	child_sa->child_id = child_id++;

	ikev2_insert_child(ike_sa, child_sa);
	++ike_sa->child_created;

	return child_sa;
}

void
ikev2_destroy_child_sa(struct ikev2_child_sa *sa)
{
	/* ikev2_remove_child() must be called before this */

	sadb_request_finish(&sa->sadb_request);
	if (sa->local)
		rc_free(sa->local);
	if (sa->remote)
		rc_free(sa->remote);
	if (sa->dhpriv)
		vfreez(sa->dhpriv);
	if (sa->dhpub)
		vfree(sa->dhpub);
	if (sa->dhpub_p)
		vfree(sa->dhpub_p);
	if (sa->g_ir)
		vfreez(sa->g_ir);
	if (sa->n_i)
		vfree(sa->n_i);
	if (sa->n_r)
		vfree(sa->n_r);
	if (sa->ts_i)
		vfree(sa->ts_i);
	if (sa->ts_r)
		vfree(sa->ts_r);
	if (sa->selector)
		rcf_free_selector(sa->selector);
	if (sa->my_proposal)
		proplist_discard(sa->my_proposal);
	if (sa->peer_proposal)
		proppair_discard(sa->peer_proposal);
	if (sa->child_param.ts_i)
		vfree(sa->child_param.ts_i);
	if (sa->child_param.ts_r)
		vfree(sa->child_param.ts_r);

	if (sa->timer)
		SCHED_KILL(sa->timer);

	racoon_free(sa);
}

struct ikev2_child_sa *
ikev2_create_child_initiator(struct ikev2_sa *ike_sa)
{
	struct ikev2_child_sa *child_sa;

	child_sa = ikev2_create_child_sa(ike_sa);
	if (!child_sa)
		return 0;

	child_sa->is_initiator = TRUE;

	return child_sa;

	/* fail:
	 *   return 0;
	 */
}

/*
 * creates a responder child_sa
 * then issues GETSPI
 */
struct ikev2_child_sa *
ikev2_create_child_responder(struct ikev2_sa *ike_sa,
			     struct sockaddr *local,
			     struct sockaddr *remote,
			     u_int32_t request_message_id,
			     struct ikev2_payload_header *sa_payload,
			     struct ikev2_payload_header *proposed_ts_i,
			     struct ikev2_payload_header *proposed_ts_r,
			     vchar_t *g_i,
			     vchar_t *n_i,
			     struct ikev2_child_param *child_param,
			     int is_createchild)
{
	size_t nonce_size;
	struct prop_pair **parsed_sa = 0;
	struct rcf_selector *sel;
	struct rcf_policy *pol;
	struct prop_pair **my_proposal = 0;
	struct prop_pair *matching_peer_proposal = 0;
	struct prop_pair *matching_my_proposal = 0;
	struct ikev2_child_sa *child_sa = 0;
	vchar_t *dhpriv = 0;
	int lifetime;
	extern struct sadb_request_method sadb_responder_request_method;

	TRACE((PLOGLOC,
	       "ikev2_create_child_responder(%p, 0x%08x, %p, %p, %p, %p, %p)\n",
	       ike_sa, request_message_id, sa_payload, proposed_ts_i,
	       proposed_ts_r, n_i, child_param));
	assert(!child_param->ts_i && !child_param->ts_r);	/* these are output parameters */

	parsed_sa = ikev2_parse_sa(&ikev2_createchild_doi, sa_payload);
	if (!parsed_sa)
		goto invalid_syntax;

	child_sa = ikev2_create_child_sa(ike_sa);
	if (!child_sa)
		goto fail_nomem;
	TRACE((PLOGLOC, "child_sa: %p\n", child_sa));
	child_sa->state = IKEV2_CHILD_STATE_IDLING;
	child_sa->message_id = request_message_id;
	child_sa->child_param = *child_param;

	child_sa->local = rcs_sadup(local);
	child_sa->remote = rcs_sadup(remote);

#ifdef ENABLE_NATT
	if (child_sa->parent->local)
		rc_free(child_sa->parent->local);
	if (child_sa->parent->remote)
		rc_free(child_sa->parent->remote);
	child_sa->parent->local = rcs_sadup(local);
	child_sa->parent->remote = rcs_sadup(remote);
#endif

	/* choose conf by ts_i and ts_r, use_transport_mode */
	/* and obtain matching ts_i and ts_r in child_param */
	sel = ike_conf_find_ikev2sel_by_ts(proposed_ts_i, proposed_ts_r,
					   &child_sa->child_param,
					   ike_sa->rmconf);
	if (!sel) {
		/* additional_ts_possible? */
		/* single_pair_required? */
		goto ts_unacceptable;
	}
	child_sa->selector = sel;

	pol = sel->pl;		/* policy */
	assert(pol != 0);

	/* choose SA as a union of  sa_i2, use_transport_mode, conf */

#ifdef notyet
	for each
		NOTIFY payload USE_TRANSPORT_MODE {
		spi = ikev2_notify_data(payl);
		....;
		}
#else
	if (ike_ipsec_mode(pol) != (child_sa->child_param.use_transport_mode
				    ? RCT_IPSM_TRANSPORT : RCT_IPSM_TUNNEL)) {
		TRACE((PLOGLOC, "ipsec_mode mismatch\n"));
		if (ikev2_selector_check(ike_sa->rmconf) == RCT_PCT_EXACT)
			goto ipsec_mode_mismatch;
		TRACE((PLOGLOC, "obeying peer request (%s)\n",
		       child_sa->child_param.use_transport_mode ?
		       "transport" : "tunnel"));
	}
#endif

	my_proposal = ikev2_ipsec_conf_to_proplist(child_sa, is_createchild);
	if (!my_proposal)
		goto fail_create_proposal;

	matching_peer_proposal = ikev2_find_match(my_proposal, parsed_sa, PEER);
	if (!matching_peer_proposal)
		goto no_proposal_chosen;

	/* make duplicate to allocate my SPI */
	matching_my_proposal = ikev2_find_match(my_proposal, parsed_sa, MINE);
	if (!matching_my_proposal)
		goto no_proposal_chosen;

	if (g_i) {
		struct prop_pair *prop;
		struct ikev2transform *transf;
		struct algdef *dhdef;

		/* (draft-17)
		 * KEYMAT = prf+(SK_d, g^ir (new) | Ni | Nr )
		 */
		prop = ikev2_prop_find(matching_my_proposal,
				       IKEV2TRANSFORM_TYPE_DH);
		if (!prop)
			goto no_proposal_chosen;	/* ??? */
		if (!prop->trns)
			goto fail_internal;
		transf = (struct ikev2transform *)prop->trns;
		dhdef = ikev2_dhinfo(get_uint16(&transf->transform_id));
		if (!dhdef)
			goto fail_internal;	/* shouldn't happen */

		child_sa->dhgrp = dhdef;

		if (oakley_dh_generate((struct dhgroup *)dhdef->definition,
				       &child_sa->dhpub, &dhpriv) != 0) {
			TRACE((PLOGLOC, "failed dh_generate\n"));
			goto fail_internal;
		}
		if (oakley_dh_compute((struct dhgroup *)dhdef->definition,
				      child_sa->dhpub_p, dhpriv,
				      g_i, &child_sa->g_ir) != 0) {
			TRACE((PLOGLOC, "failed dh_compute\n"));
			goto fail_internal;
		}
	} else {		/* if (! g_i) */
		if (is_createchild &&
		    ikev2_need_pfs(ike_sa->rmconf) == RCT_BOOL_ON) {
			isakmp_log(ike_sa, local, remote, 0, PLOG_INTERR,
				   PLOGLOC, "message lacks KEi payload\n");
			++isakmpstat.malformed_message;
			goto fail;
		}
	}

	if (n_i) {
		child_sa->n_i = n_i;
		nonce_size = ikev2_nonce_size(ike_sa->rmconf);
		child_sa->n_r = random_bytes(nonce_size);
		if (!child_sa->n_r)
			goto fail_nomem;
	}

	/* save my proposal list to keep SPI values */
	child_sa->my_proposal = proplist_new();
	if (!child_sa->my_proposal)
		goto fail_nomem;
	child_sa->my_proposal[1] = matching_my_proposal;
	matching_my_proposal = 0;

	child_sa->peer_proposal = matching_peer_proposal;
	matching_peer_proposal = 0;

	/* XXX generate policy */
	if (pol->peers_sa_ipaddr && rcs_is_addr_rw(pol->peers_sa_ipaddr)) {
		IPSEC_CONF(lifetime, pol->ips, ipsec_sa_lifetime_time, 0);
		if (ike_spmif_post_policy_add(child_sa->selector,
					      ike_ipsec_mode(pol), lifetime,
					      child_sa->local, child_sa->remote,
					      ike_sa->rmconf) < 0)
			goto fail_internal;
	}

	sadb_request_initialize(&child_sa->sadb_request,
				debug_pfkey
				? &sadb_debug_method
				: &sadb_responder_request_method,
				sadb_new_seq(),
				ikev2_child_getspi_response,
				ikev2_expired, child_sa);

	TRACE((PLOGLOC, "calling getspi\n"));
	ikev2_child_getspi(child_sa);
	TRACE((PLOGLOC, "done\n"));

	/* ikev2_create_child_responder_cont() is called when
	 * state transits to GETSPI_DONE */

      done:
	if (dhpriv)
		vfreez(dhpriv);
	if (my_proposal)
		proplist_discard(my_proposal);
	if (parsed_sa)
		proplist_discard(parsed_sa);
	return child_sa;

      fail:
	if (child_sa)
		ikev2_child_state_set(child_sa, IKEV2_CHILD_STATE_EXPIRED);
	if (matching_my_proposal)
		proppair_discard(matching_my_proposal);
	if (matching_peer_proposal)
		proppair_discard(matching_peer_proposal);
	child_sa = 0;
	goto done;

      invalid_syntax:
	isakmp_log(ike_sa, 0, 0, 0,
		   PLOG_PROTOERR, PLOGLOC, "invalid SA payload syntax\n");
	goto fail;
      ipsec_mode_mismatch:
	isakmp_log(ike_sa, 0, 0, 0,
		   PLOG_PROTOERR, PLOGLOC,
		   "ipsec mode does not match.  exact match specified\n");
	goto fail;
      no_proposal_chosen:
	isakmp_log(ike_sa, 0, 0, 0,
		   PLOG_PROTOERR, PLOGLOC, "no proposal chosen\n");
	goto fail;
      fail_create_proposal:
	isakmp_log(ike_sa, 0, 0, 0,
		   PLOG_PROTOERR, PLOGLOC, "failed creating proposal\n");
	goto fail;
      ts_unacceptable:
	isakmp_log(ike_sa, 0, 0, 0,
		   PLOG_PROTOERR, PLOGLOC, "ts unacceptable\n");
	goto fail;
      fail_nomem:
	isakmp_log(ike_sa, 0, 0, 0,
		   PLOG_INTERR, PLOGLOC, "failed allocating memory\n");
	goto fail;
      fail_internal:
	isakmp_log(ike_sa, 0, 0, 0,
		   PLOG_INTERR, PLOGLOC, "failed for internal error\n");
	goto fail;
}

/*
 * ikev2_create_child_responder_cont:
 *     called when child_sa state transits from GETSPI to MATURE
 */
static void
ikev2_create_child_responder_cont(struct ikev2_child_sa *child_sa)
{
	struct ikev2_sa *ike_sa;

	TRACE((PLOGLOC, "ikev2_create_child_responder_cont(%p)\n", child_sa));

	assert(!child_sa->is_initiator);
	assert(child_sa->parent != 0);
	ike_sa = child_sa->parent;

	ikev2_add_ipsec_sa(child_sa, &child_sa->child_param,
			   child_sa->peer_proposal, child_sa->my_proposal[1]);

	/* #if defined(__FreeBSD__) || defined(__NetBSD__) */
	/* KAME does not generate hard lifetime expiration message */
	/* start expiration timer */
	{
		struct rcf_ipsec *conf;
		int lifetime;

		conf = child_sa->selector->pl->ips;
		IPSEC_CONF(lifetime, conf, ipsec_sa_lifetime_time, 0);
		if (lifetime) {
			child_sa->timer =
				sched_new(lifetime, ikev2_child_expire_callback,
					  child_sa);
			if (!child_sa->timer) {
				isakmp_log(ike_sa, 0, 0, 0,
					   PLOG_INTERR, PLOGLOC,
					   "failed allocating memory\n");
				ikev2_child_state_set(child_sa,
						      IKEV2_CHILD_STATE_EXPIRED);
			}
		}
	}
	/* #endif */

	TRACE((PLOGLOC, "ike_sa state %d\n", ike_sa->state));
	switch (ike_sa->state) {
	case IKEV2_STATE_RES_IKE_SA_INIT_SENT:
		ikev2_responder_state1_send(ike_sa, child_sa);
		break;
	case IKEV2_STATE_ESTABLISHED:
		ikev2_createchild_responder_send(ike_sa, child_sa);
		break;
	default:
		/* unexpected */
		isakmp_log(ike_sa, 0, 0, 0,
			   PLOG_INTERR, PLOGLOC,
			   "unexpected state %d\n", ike_sa->state);
		break;
	}
}

void
ikev2_insert_child(struct ikev2_sa *ike_sa, struct ikev2_child_sa *child_sa)
{
	IKEV2_CHILD_LIST_LINK(&ike_sa->children, child_sa);
	child_sa->parent = ike_sa;
}

void
ikev2_remove_child(struct ikev2_child_sa *child_sa)
{
	IKEV2_CHILD_LIST_REMOVE(&child_sa->parent->children, child_sa);
}

struct ikev2_child_sa *
ikev2_choose_pending_child(struct ikev2_sa *ike_sa, int may_be_informational)
{
	struct ikev2_child_sa *sa;

#ifdef notyet
	/* window */
#else
	if (ike_sa->request_pending > 0)
		return 0;
#endif

	for (sa = IKEV2_CHILD_LIST_FIRST(&ike_sa->children);
	     sa;
	     sa = IKEV2_CHILD_LIST_NEXT(sa)) {
		if (sa->is_initiator
		    && sa->state == IKEV2_CHILD_STATE_GETSPI_DONE)
			return sa;
		if (may_be_informational
		    && sa->is_initiator
		    && sa->state == IKEV2_CHILD_STATE_REQUEST_PENDING)
			return sa;
	}
	return 0;
}

struct ikev2_child_sa *
ikev2_find_child_sa(struct ikev2_sa *ike_sa, int is_responder, u_int32_t id)
{
	struct ikev2_child_sa *sa;

	for (sa = IKEV2_CHILD_LIST_FIRST(&ike_sa->children);
	     !IKEV2_CHILD_LIST_END(sa);
	     sa = IKEV2_CHILD_LIST_NEXT(sa)) {
		if (((is_responder && !sa->is_initiator)
		     || (!is_responder && sa->is_initiator))
		    && sa->message_id == id)
			return sa;
	}
	return 0;
}

struct ikev2_child_sa *
ikev2_find_request(struct ikev2_sa *ike_sa, u_int32_t id)
{
	struct ikev2_child_sa	* sa;

	for (sa = IKEV2_CHILD_LIST_FIRST(&ike_sa->children);
	     !IKEV2_CHILD_LIST_END(sa);
	     sa = IKEV2_CHILD_LIST_NEXT(sa)) {
		if (sa->is_initiator
		    && sa->message_id == id
		    && (sa->state == IKEV2_CHILD_STATE_WAIT_RESPONSE
			|| sa->state == IKEV2_CHILD_STATE_REQUEST_SENT))
			return sa;
	}
	return 0;
}


struct ikev2_child_sa *
ikev2_find_child_by_id(struct ikev2_sa *ike_sa, u_int id)
{
	struct ikev2_child_sa *sa;
	for (sa = IKEV2_CHILD_LIST_FIRST(&ike_sa->children);
	     !IKEV2_CHILD_LIST_END(sa);
	     sa = IKEV2_CHILD_LIST_NEXT(sa)) {
		if (sa->child_id == id)
			return sa;
	}
	return 0;
}

struct ikev2_child_sa *
ikev2_find_child_sa_by_spi(struct ikev2_sa *ike_sa,
			   u_int protocol_id, u_int32_t spi,
			   enum peer_mine which)
{
	struct ikev2_child_sa *child_sa;
	struct prop_pair *proposal;

	for (child_sa = IKEV2_CHILD_LIST_FIRST(&ike_sa->children);
	     !IKEV2_CHILD_LIST_END(child_sa);
	     child_sa = IKEV2_CHILD_LIST_NEXT(child_sa)) {
		if (child_sa->state != IKEV2_CHILD_STATE_MATURE)
			continue;

		proposal = (which == PEER) ?
		    child_sa->peer_proposal :
		    child_sa->my_proposal[1];
		assert(proposal != 0);
		for (; proposal; proposal = proposal->next) {
			struct isakmp_pl_p *prop;
			prop = proposal->prop;
			TRACE((PLOGLOC, "proto_id %d spi 0x%08x\n", prop->proto_id, get_uint32((u_int32_t *)(prop + 1))));
			if (prop->proto_id == protocol_id
			    && get_uint32((u_int32_t *)(prop + 1)) == spi)
				return child_sa;
		}
	}
	return 0;
}

void
ikev2_wakeup_child_sa(struct ikev2_child_sa *child_sa)
{
	struct ikev2_sa *ike_sa;

	assert(child_sa != 0);
	ike_sa = child_sa->parent;
	switch (child_sa->state) {
	case IKEV2_CHILD_STATE_GETSPI_DONE:
		ikev2_createchild_initiator_send(ike_sa, child_sa);
		break;
	case IKEV2_CHILD_STATE_REQUEST_PENDING:
		if (!child_sa->callback) {
			isakmp_log(ike_sa, 0, 0, 0,
				   PLOG_INTERR, PLOGLOC,
				   "shouldn't happen: no callback\n");
			ikev2_child_state_set(child_sa,
					      IKEV2_CHILD_STATE_EXPIRED);
			break;
		}
		child_sa->callback(REQUEST_CALLBACK_CONTINUE, child_sa,
				   child_sa->callback_param);
		break;
	default:
		isakmp_log(ike_sa, 0, 0, 0,
			   PLOG_INTERR, PLOGLOC,
			   "shouldn't happen: unexpected child_sa state %d\n",
			   child_sa->state);
		break;
	}
}

static int
inet_addr_bits(struct sockaddr *a)
{
	switch (a->sa_family) {
	case AF_INET:
		return 32;
		break;
#ifdef INET6
	case AF_INET6:
		return 128;
		break;
#endif
	default:
		isakmp_log(0, 0, 0, 0,
			   PLOG_INTERR, PLOGLOC,
			   "unsupported address family (type %d)\n",
			   a->sa_family);
		return 0;
		break;
	}
}

int
ikev2_child_getspi(struct ikev2_child_sa *child_sa)
{
	struct rcpfk_msg param;
	struct prop_pair *proto;
	struct rc_addrlist *src, *dst;

	assert(child_sa->state == IKEV2_CHILD_STATE_IDLING);
	assert(child_sa->my_proposal[1] != 0);

	ikev2_child_state_set(child_sa, IKEV2_CHILD_STATE_GETSPI);

	for (proto = child_sa->my_proposal[1]; proto; proto = proto->next) {
		if (get_uint32((u_int32_t *)(proto->prop + 1)) != 0) {
			/* user specified spi? */
			u_int32_t spi =
				get_uint32((u_int32_t *)(proto->prop + 1));
			isakmp_log(child_sa->parent, 0, 0, 0, PLOG_DEBUG,
				   PLOGLOC, "using specified spi %d (0x%x)\n",
				   spi, spi);
			continue;
		}

		switch (proto->prop->proto_id) {
		case IKEV2PROPOSAL_ESP:
			param.satype = RCT_SATYPE_ESP;
			break;
		case IKEV2PROPOSAL_AH:
			param.satype = RCT_SATYPE_AH;
			break;
		default:
			goto unexpected_proto_id;
		}
		param.seq = child_sa->sadb_request.seqno;

		/* for X_EXT_SA2 */
		assert(child_sa->selector->pl);
		param.samode = ike_ipsec_mode(child_sa->selector->pl);
		param.reqid = child_sa->selector->reqid;

		/* for SRC */
		src = child_sa->selector->pl->peers_sa_ipaddr;
		if (src && !rcs_is_addr_rw(src)) {
			if (src->type == RCT_ADDR_INET) {
				param.sa_src = src->a.ipaddr;
				param.pref_src = addr_prefixlen(src);
			} else {
				isakmp_log(child_sa->parent, 0, 0, 0,
					   PLOG_INTERR, PLOGLOC,
					   "peers_sa_ipaddr is unsupported address type (type %s)\n",
					   rct2str(src->type));
				return -1;
			}
		} else {
			param.sa_src = child_sa->parent->remote;
			param.pref_src = inet_addr_bits(param.sa_src);
		}

		/* for DST */
		dst = child_sa->selector->pl->my_sa_ipaddr;
		if (dst) {
			if (dst->type == RCT_ADDR_INET) {
				param.sa_dst = dst->a.ipaddr;
				param.pref_dst = addr_prefixlen(dst);
			} else {
				isakmp_log(child_sa->parent, 0, 0, 0,
					   PLOG_INTERR, PLOGLOC,
					   "my_sa_ipaddr is unsupported address type (type %s)\n",
					   rct2str(dst->type));
				return -1;
			}
		} else {
			param.sa_dst = child_sa->parent->local;
			param.pref_dst = inet_addr_bits(param.sa_dst);
		}

		param.ul_proto = child_sa->selector->upper_layer_protocol;

		if (child_sa->sadb_request.method->getspi(&param) != 0) {
			isakmp_log(child_sa->parent, 0, 0, 0,
				   PLOG_INTERR, PLOGLOC,
				   "failed sending getspi\n");
			/* continue? */
		}
	}
	return 0;

      unexpected_proto_id:
	isakmp_log(child_sa->parent, 0, 0, 0,
		   PLOG_INTERR, PLOGLOC,
		   "internal error: unexpected proto_id (%d)\n",
		   proto->prop->proto_id);
	return -1;
}

/*ARGSUSED*/
int
ikev2_child_getspi_response(struct sadb_request *req,
			    struct sockaddr *src, struct sockaddr *dst,
			    u_int satype, u_int32_t spi)
{
	struct ikev2_child_sa *child_sa;
	struct prop_pair *proposal;
	int proto;
	int found;
	int not_completed;

	child_sa = req->sa;
	assert(child_sa != 0);

	/*
	 * for each proposal,
	 * if satype matches, assign SPI
	 * if no corresponding proposal
	 * return -1;
	 */

	switch (satype) {
	case RCT_SATYPE_ESP:
		proto = IKEV2PROPOSAL_ESP;
		break;
	case RCT_SATYPE_AH:
		proto = IKEV2PROPOSAL_AH;
		break;
	default:
		goto invalid_param;
	}

	assert(child_sa->my_proposal[1] != 0);	/* only one proposal is generated from config,
						 * and it is in my_proposal[1] */

	found = FALSE;
	not_completed = FALSE;
	for (proposal = child_sa->my_proposal[1];
	     proposal;
	     proposal = proposal->next) {
		struct isakmp_pl_p *prop;
		assert(proposal->prop != 0);
		prop = proposal->prop;
		if (!found && prop->proto_id == proto) {
			assert(get_uint16(&prop->h.len) ==
			       sizeof(struct isakmp_pl_p) + sizeof(u_int32_t));
			found = TRUE;
			put_uint32((u_int32_t *)(prop + 1), spi);
		}
		if (get_uint32((u_int32_t *)(prop + 1)) == 0)
			not_completed = TRUE;
	}

	if (!found) {
		/* log it */
		isakmp_log(child_sa->parent, 0, 0, 0,
			   PLOG_INTERR, PLOGLOC,
			   "can't find correspoonding %s proposal\n",
			   satype == RCT_SATYPE_ESP ? "ESP" : "AH");
		return -1;
	}

	if (not_completed)
		return 0;

	ikev2_child_state_next(child_sa);

	return 0;

      invalid_param:
	isakmp_log(child_sa->parent, 0, 0, 0,
		   PLOG_INTERR, PLOGLOC, "invalid value returned from pfkey\n");
	return -1;
}

/*
 * ikev2_child_getspi_done:
 *     called when initiator child_sa state changes to GETSPI_DONE
 */
static void
ikev2_child_getspi_done(struct ikev2_child_sa *child_sa)
{
	struct ikev2_sa *ike_sa;

	assert(child_sa->is_initiator);
	assert(child_sa->parent != 0);
	ike_sa = child_sa->parent;

	switch (ike_sa->state) {
	case IKEV2_STATE_IDLING:
		assert(ike_sa->is_initiator);
		ikev2_initiator_start(ike_sa);
		break;
	case IKEV2_STATE_ESTABLISHED:
		child_sa = ikev2_choose_pending_child(ike_sa, TRUE);
		if (child_sa)
			ikev2_wakeup_child_sa(child_sa);
		break;
	case IKEV2_STATE_DYING:
	case IKEV2_STATE_DEAD:
		/* parent expired while the child was in GETSPI state? */
		/* send error */
		isakmp_log(child_sa->parent, 0, 0, 0,
			   PLOG_INTERR, PLOGLOC,
			   "ike_sa expired while child is in GETSPI state\n");
		ikev2_child_abort(child_sa, ETIMEDOUT);
		break;
	default:
		/* probably multiple negotiations are going on */
		/* wait until ikev2_choose_pending_child() is called */
		TRACE((PLOGLOC, "%p pending\n", child_sa));
		break;
	}
	return;
}

static int
ikev2_sadb_update(struct ikev2_child_sa *child_sa,
		  int (*update_func) (struct rcpfk_msg *),
		  struct rcpfk_msg *param, void *data)
{
	u_int8_t **keypp = (u_int8_t **)data;
	int err;
	struct rcf_ipsec *conf;

	/*
	 * (draft-17)
	 * If multiple IPsec protocols are negotiated, keying material is
	 * taken in the order in which the protocol headers will appear in
	 * the encapsulated packet.
	 * 
	 * If a single protocol has both encryption and authentication keys,
	 * the encryption key is taken from the first octets of KEYMAT and
	 * the authentication key is taken from the next octets.
	 */

	/* 
	 * (draft-17)
	 tunnel encapsulators and
	 decapsulators for all tunnel-mode Security Associations (SAs) created
	 by IKEv2 MUST support the ECN full-functionality option for tunnels
	 specified in [RFC3168] and MUST implement the tunnel encapsulation
	 and decapsulation processing specified in [RFC2401bis] to prevent
	 discarding of ECN congestion indications.
	 */
	/* thus don't specify SADB_SAFLAGS_NOECN */

	param->saflags = 0;

	param->wsize = ikev2_ipsec_window_size;

	conf = child_sa->selector->pl->ips;
	IPSEC_CONF(param->lft_hard_time, conf, ipsec_sa_lifetime_time, 0);
#ifdef notyet
	IPSEC_CONF(param->lft_hard_bytes, conf, ipsec_sa_lifetime_byte, 0);
#else
	param->lft_hard_bytes = 0;
#endif
	param->lft_soft_time = param->lft_hard_time *
		(ikev2_lifetime_soft_factor +
		 ikev2_lifetime_soft_jitter * ((double)eay_random_uint32() /
					       UINT32_MAX));
	param->lft_soft_bytes =
		param->lft_hard_bytes * (ikev2_lifetime_soft_factor +
					 ikev2_lifetime_soft_jitter *
					 ((double)eay_random_uint32() /
					  UINT32_MAX));

	param->enckey = (caddr_t)*keypp;
	*keypp += param->enckeylen;
	param->authkey = (caddr_t)*keypp;
	*keypp += param->authkeylen;

	err = update_func(param);	/* update or add */

	return err;
}

static struct sockaddr *
determine_sa_endpoint(struct ikev2_child_sa *child_sa,
    struct sockaddr_storage *ss,
    struct rc_addrlist *config_ipaddr, struct sockaddr *actual_addr)
{
	struct sockaddr *addr;
	int is_natt;

#ifdef ENABLE_NATT
	is_natt = child_sa->parent->behind_nat ||
		  child_sa->parent->peer_behind_nat;
#else
	is_natt = 0;
#endif
	if (config_ipaddr) {
		if (config_ipaddr->type == RCT_ADDR_INET) {
#ifdef ENABLE_NATT
			if (is_natt) {
				memcpy(ss, config_ipaddr->a.ipaddr,
				       SOCKADDR_LEN(config_ipaddr->a.ipaddr));
				addr = (struct sockaddr *)ss;
				if (!set_port(addr,
					      extract_port(actual_addr))) {
					isakmp_log(child_sa->parent, 0, 0, 0,
						   PLOG_INTERR, PLOGLOC,
						   "set_port for NAT-T failed\n");
					return NULL;
				}
			} else
				addr = config_ipaddr->a.ipaddr;
#else
			addr = config_ipaddr->a.ipaddr;
#endif
		} else {
			isakmp_log(child_sa->parent, 0, 0, 0,
				   PLOG_INTERR, PLOGLOC,
				   "my_sa_ipaddr or peers_sa_ipaddr is "
				   "unsupported address type (type %s)\n",
				   rct2str(config_ipaddr->type));
			return NULL;
		}
	} else {
		if (is_natt)
			addr = actual_addr;
		else {
			memcpy(ss, actual_addr, SOCKADDR_LEN(actual_addr));
			addr = (struct sockaddr *)ss;
#if 0
			rcs_setsaport(addr, 0);
#endif
		}
	}
	return addr;
}

static int
ikev2_sadb_outbound(struct ikev2_child_sa *child_sa, struct rcpfk_msg *param,
		    void *data)
{
	struct rcf_policy *p;
	struct sockaddr *my_addr;
	struct sockaddr *peer_addr;
	struct sockaddr_storage my_ss, peer_ss;

	p = child_sa->selector->pl;	/* policy */

	my_addr = determine_sa_endpoint(child_sa, &my_ss,
					p->my_sa_ipaddr,
					child_sa->parent->local);
	if (my_addr == NULL)
		return -1;

	peer_addr = determine_sa_endpoint(child_sa, &peer_ss,
					  p->peers_sa_ipaddr &&
					  !rcs_is_addr_rw(p->peers_sa_ipaddr) ?
					  p->peers_sa_ipaddr : NULL,
					  child_sa->parent->remote);
	if (peer_addr == NULL)
		return -1;

	param->sa_src = my_addr;
	param->sa_dst = peer_addr;

	/* KAME/USAGI PF_KEY checks prefixes eventhough it does not use
	 * them for SADB_ADD/UPDATE.  we need to fill with some sane value
	 * here
	 */
	param->pref_src = 0;
	param->pref_dst = 0;

	return ikev2_sadb_update(child_sa,
				 child_sa->sadb_request.method->add_outbound,
				 param, data);
}

static int
ikev2_sadb_inbound(struct ikev2_child_sa *child_sa, struct rcpfk_msg *param,
		   void *data)
{
	struct rcf_policy *p;
	struct sockaddr *my_addr;
	struct sockaddr *peer_addr;
	struct sockaddr_storage my_ss, peer_ss;

	p = child_sa->selector->pl;

	my_addr = determine_sa_endpoint(child_sa, &my_ss,
					p->my_sa_ipaddr,
					child_sa->parent->local);
	if (my_addr == NULL)
		return -1;

	peer_addr = determine_sa_endpoint(child_sa, &peer_ss,
					  p->peers_sa_ipaddr &&
					  !rcs_is_addr_rw(p->peers_sa_ipaddr) ?
					  p->peers_sa_ipaddr : NULL,
					  child_sa->parent->remote);
	if (peer_addr == NULL)
		return -1;

	param->sa_src = peer_addr;
	param->sa_dst = my_addr;
	param->pref_src = 0;
	param->pref_dst = 0;

	param->seq = child_sa->sadb_request.seqno;

	return ikev2_sadb_update(child_sa,
				 child_sa->sadb_request.method->update_inbound,
				 param, data);
}

static int
calculate_keylen(struct ikev2_child_sa *child_sa, struct rcpfk_msg *param,
		 void *data)
{
	int **keylen = (int **)data;

	*keylen += param->enckeylen + param->authkeylen;
	return 0;
}

static int
ikev2_add_ipsec_sa(struct ikev2_child_sa *child_sa,
		   struct ikev2_child_param *child_param,
		   struct prop_pair *matching_peer_proposal,
		   struct prop_pair *matching_my_proposal)
{
	int err = 0;
	int required_len;
	u_int8_t *keyptr;
	vchar_t *keymat = 0;

	/* scan the proposal tree to calculate required key length */
	required_len = 0;
	if (ikev2_proposal_to_ipsec(child_sa, child_param, matching_my_proposal,
				    calculate_keylen, &required_len)) {
		err = IKEV2_NO_PROPOSAL_CHOSEN;
		goto bailout;
	}

	/* then compute keymat.  "2 *" for inbound and outbound */
	keymat = compute_keymat(child_sa->parent, child_sa->g_ir,
				2 * required_len, child_sa->n_i, child_sa->n_r);
	if (!keymat) {
		err = -1;	/* ??? */
		goto bailout;
	}

	/*
	 * call sequence:
	 * ikev2_proposal_to_ipsec()
	 *   -> ikev2_sadb_inbound()/outbound()
	 *     -> ikev2_sadb_update()
	 *       -> sadb_add() / sadb_update()
	 */

	/* (draft-17)
	 * All keys for SAs carrying data from the initiator to the responder
	 * are taken before SAs going in the reverse direction.
	 */
	keyptr = (u_int8_t *)keymat->v;
	if (ikev2_proposal_to_ipsec(child_sa, child_param,
				    (child_sa->is_initiator ? matching_peer_proposal :
				     matching_my_proposal),
				    (child_sa->is_initiator ? ikev2_sadb_outbound :
				     ikev2_sadb_inbound),
				    &keyptr)) {
		err = IKEV2_NO_PROPOSAL_CHOSEN;
		goto bailout;
	}

	if (ikev2_proposal_to_ipsec(child_sa, child_param,
				    (child_sa->is_initiator ? matching_my_proposal :
				     matching_peer_proposal),
				    (child_sa->is_initiator ? ikev2_sadb_inbound :
				     ikev2_sadb_outbound),
				    &keyptr)) {
		err = IKEV2_NO_PROPOSAL_CHOSEN;
		goto bailout;
	}

      bailout:
	if (keymat)
		vfreez(keymat);
	return err;
}

/*
 * update the ipsec SA of initiator child_sa
 */
void
ikev2_update_child(struct ikev2_child_sa *child_sa,
		   struct ikev2_payload_header *sa_r2,
		   struct ikev2_payload_header *ts_i,
		   struct ikev2_payload_header *ts_r,
		   struct ikev2_child_param *param)
{
	struct prop_pair **parsed_sa;
	struct prop_pair *matching_proposal = 0;
	struct prop_pair *matching_my_proposal = 0;
	struct prop_pair **new_my_proposal_list = 0;
	vchar_t *g_ir;
	int err = 0;

	/* update IPsec SA with received parameter */

	parsed_sa = ikev2_parse_sa(&ikev2_createchild_doi, sa_r2);
	if (!parsed_sa) {
		isakmp_log(child_sa->parent, 0, 0, 0,
			   PLOG_PROTOERR, PLOGLOC,
			   "failed to parse SA payload\n");
		err = IKEV2_INVALID_SYNTAX;
		goto abort;
	}

	matching_proposal =
		ikev2_find_match(child_sa->my_proposal, parsed_sa, PEER);
	if (!matching_proposal) {
		isakmp_log(child_sa->parent, 0, 0, 0,
			   PLOG_PROTOERR, PLOGLOC,
			   "peer response does not match my proposal\n");
		err = IKEV2_NO_PROPOSAL_CHOSEN;
		goto abort;
	}

	/* find match to copy my SPI */
	matching_my_proposal =
		ikev2_find_match(child_sa->my_proposal, parsed_sa, MINE);
	if (!matching_my_proposal) {
		/* shouldn't happen */
		isakmp_log(child_sa->parent, 0, 0, 0,
			   PLOG_INTERR, PLOGLOC,
			   "shouldn't happen: failed ikev2_find_match\n");
		err = -1;
		goto abort;
	}

	/* confirm TSi and TSr do not contradict with my proposal */
	switch (ikev2_confirm_ts(ts_i, ts_r, child_sa->selector)) {
	case -1:
		isakmp_log(child_sa->parent, 0, 0, 0,
			   PLOG_PROTOERR, PLOGLOC,
			   "responder's TSi does not match my selector\n");
		err = IKEV2_NO_PROPOSAL_CHOSEN;
		goto abort;
	case -2:
		isakmp_log(child_sa->parent, 0, 0, 0,
			   PLOG_PROTOERR, PLOGLOC,
			   "responder's TSr does not match my selector\n");
		err = IKEV2_NO_PROPOSAL_CHOSEN;
		goto abort;
	default:
		break;
	}

	/*
	 * child_sa->src = ts_to_addr(ts_i);
	 * child_sa->dst = ts_to_addr(ts_r);
	 */

	if (ike_ipsec_mode(child_sa->selector->pl) != (param->use_transport_mode
						       ? RCT_IPSM_TRANSPORT
						       : RCT_IPSM_TUNNEL)) {
		TRACE((PLOGLOC, "ipsec_mode mismatch\n"));
		if (ikev2_selector_check(child_sa->parent->rmconf) == RCT_PCT_EXACT) {
			isakmp_log(child_sa->parent, 0, 0, 0, PLOG_PROTOERR,
				   PLOGLOC, "mode mismatch: peer %s mine %s\n",
				   (param->use_transport_mode ? "transport" :
				    "tunnel"),
				   (ike_ipsec_mode(child_sa->selector->pl) ==
				    RCT_IPSM_TRANSPORT ? "transport" :
				    "tunnel"));
			goto abort;
		}
		TRACE((PLOGLOC, "obeying peer request (%s)\n",
		       child_sa->child_param.
		       use_transport_mode ? "transport" : "tunnel"));
	}

	g_ir = 0;
#ifdef notyet
	/* if (ke_i && ke_r) g_ir = g^i^r */
#endif

	/* replace my_proposal with matching_my_proposal */
	new_my_proposal_list = proplist_new();
	if (!new_my_proposal_list)
		goto abort_nomem;
	new_my_proposal_list[1] = matching_my_proposal;
	matching_my_proposal = 0;
	proplist_discard(child_sa->my_proposal);
	child_sa->my_proposal = new_my_proposal_list;
	new_my_proposal_list = 0;

	/* remember peer proposal */
	assert(child_sa->peer_proposal == 0);
	child_sa->peer_proposal = matching_proposal;
	matching_proposal = 0;

	err = ikev2_add_ipsec_sa(child_sa, param, child_sa->peer_proposal,
				 child_sa->my_proposal[1]);
	if (err) {
		isakmp_log(child_sa->parent, 0, 0, 0,
			   PLOG_INTERR, PLOGLOC, "failed creating ipsec SA\n");
		goto abort;
	}

	ikev2_child_state_set(child_sa, IKEV2_CHILD_STATE_MATURE);

	/* #if defined(__FreeBSD__) || defined(__NetBSD__) */
	/* KAME does not generate hard lifetime expiration message */
	/* start expiration timer */
	{
		struct rcf_ipsec *conf;
		int lifetime;

		conf = child_sa->selector->pl->ips;
		IPSEC_CONF(lifetime, conf, ipsec_sa_lifetime_time, 0);
		if (lifetime) {
			child_sa->timer =
				sched_new(lifetime, ikev2_child_expire_callback,
					  child_sa);
			if (!child_sa->timer)
				goto abort_nomem;
		}
	}
	/* #endif */

      done:
	if (new_my_proposal_list)
		proplist_discard(new_my_proposal_list);
	if (matching_my_proposal)
		proppair_discard(matching_my_proposal);
	if (matching_proposal)
		proppair_discard(matching_proposal);
	if (parsed_sa)
		proplist_discard(parsed_sa);
	return;

      abort_nomem:
	isakmp_log(child_sa->parent, 0, 0, 0,
		   PLOG_INTERR, PLOGLOC, "failed allocating memory\n");
      abort:
	ikev2_child_abort(child_sa, ECONNREFUSED);	/* ??? */

#if 0				/* not sure whether this is correct */
	int i;

	/* notify err */
	if (err <= 0)
		err = IKEV2_INVALID_SYNTAX;
	for (i = 0; i < 255; ++i) {
		struct prop_pair *proposal;

		proposal = parsed_sa[i];
		if (proposal) {
			struct isakmp_pl_p *prop;
			u_int proto;
			u_int8_t *spi;
			size_t spi_size;
			struct ikev2_payloads *payl;

			prop = proposal->prop;
			assert(prop != 0);
			proto = prop->proto_id;
			if (proto != IKEV2PROPOSAL_ESP)
				continue;

			assert(IKEV2PROPOSAL_ESP == IKEV2_NOTIFY_PROTO_ESP);
			spi = (u_int8_t *)(prop + 1);
			spi_size = prop->spi_size;

			payl = racoon_malloc(sizeof(struct ikev2_payloads));
			ikev2_payloads_init(payl);
			ikev2_payloads_push(payl, IKEV2_PAYLOAD_NOTIFY,
					    ikev2_delete_payload(proto,
								 spi_size, 1,
								 spi),
					    TRUE);
			informational_initiator_notify(child_sa->parent, payl);
			break;
		}
	}
#endif
	goto done;
}

/* #if defined(__FreeBSD__) || defined(__NetBSD__) */
/*
 * timer callback for child_sa expiration
 */
static void
ikev2_child_expire_callback(void *param)
{
	struct ikev2_child_sa *child_sa;

	child_sa = (struct ikev2_child_sa *)param;
	SCHED_KILL(child_sa->timer);
	ikev2_expire_child(child_sa);
}
/* #endif */

/*ARGSUSED*/
void
ikev2_expired(struct sadb_request *req, u_int satype, u_int32_t spi,
	      struct sockaddr *src, struct sockaddr *dst,
	      u_int samode, int expired)
{
	struct ikev2_child_sa *child_sa;

	TRACE((PLOGLOC, "ikev2_expired(%p, %d, 0x%x, %p, %p, %d, %d)\n",
	       req, satype, spi, src, dst, samode, expired));

	child_sa = req->sa;
	assert(child_sa != 0);

	switch (expired) {
	case 1:		/* soft expired */
		if (!child_sa->rekey_inprogress) {
			child_sa->rekey_inprogress = TRUE;
			ikev2_rekey_childsa(child_sa, satype, spi);
		} else {
			TRACE((PLOGLOC, "rekey already in progress\n"));
		}
		return;
	case 2:		/* hard expired */
#if 1
		/*
		 * hard expire is not used, due to difference of KAME and USAGI.
		 * instead, use child_sa->timer
		 */
#else
		ikev2_expire_child(child_sa);
#endif
		break;
	default:
		TRACE((PLOGLOC, "unexpected %d\n", expired));
		return;
	}
}

static void
ikev2_expire_child(struct ikev2_child_sa *child_sa)
{
	TRACE((PLOGLOC, "expire child %p (state %d)\n", child_sa,
	       child_sa->state));
	switch (child_sa->state) {
	case IKEV2_CHILD_STATE_MATURE:
		ikev2_child_delete(child_sa);
		break;
	default:
		ikev2_child_state_set(child_sa, IKEV2_CHILD_STATE_EXPIRED);
		break;
	}
}

/*
 * issue sadb delete request
 */
void
ikev2_delete_sa(struct ikev2_child_sa *child_sa, int protocol_id,
		struct sockaddr *src, struct sockaddr *dst, u_int32_t spi)
{
	struct rcpfk_msg param;

	switch (protocol_id) {
	case IKEV2PROPOSAL_ESP:
		param.satype = RCT_SATYPE_ESP;
		break;
	case IKEV2PROPOSAL_AH:
		param.satype = RCT_SATYPE_AH;
		break;
	}
	param.sa_src = src;
	param.sa_dst = dst;
	param.spi = htonl(spi);
	(void)child_sa->sadb_request.method->delete_sa(&param);
}

void
ikev2_child_delete_outbound(struct ikev2_child_sa *child_sa)
{
	struct prop_pair *proposal;
	struct ikev2_sa *ike_sa;
	struct rcf_policy *policy;
	struct sockaddr *local;
	struct sockaddr *remote;

	ike_sa = child_sa->parent;
	policy = child_sa->selector->pl;
	local = (policy->my_sa_ipaddr ?
	    policy->my_sa_ipaddr->a.ipaddr : ike_sa->local);
	remote = (policy->peers_sa_ipaddr &&
	    !rcs_is_addr_rw(policy->peers_sa_ipaddr) ?
	    policy->peers_sa_ipaddr->a.ipaddr : ike_sa->remote);

	for (proposal = child_sa->peer_proposal;
	     proposal;
	     proposal = proposal->next) {
		struct isakmp_pl_p *prop;
		u_int32_t outbound_spi;

		prop = proposal->prop;
		outbound_spi = get_uint32((u_int32_t *)(prop + 1));
		ikev2_delete_sa(child_sa, prop->proto_id, local, remote,
				outbound_spi);
	}
}

void
ikev2_child_delete_inbound(struct ikev2_child_sa *child_sa)
{
	struct prop_pair *proposal;
	struct ikev2_sa *ike_sa;
	struct rcf_policy *policy;
	struct sockaddr *local;
	struct sockaddr *remote;

	ike_sa = child_sa->parent;
	policy = child_sa->selector->pl;
	local = (policy->my_sa_ipaddr ?
	    policy->my_sa_ipaddr->a.ipaddr : ike_sa->local);
	remote = (policy->peers_sa_ipaddr &&
	    !rcs_is_addr_rw(policy->peers_sa_ipaddr) ?
	    policy->peers_sa_ipaddr->a.ipaddr : ike_sa->remote);

	/* delete inbound */
	for (proposal = child_sa->my_proposal[1];
	     proposal;
	     proposal = proposal->next) {
		struct isakmp_pl_p *prop;
		u_int32_t inbound_spi;

		assert(proposal->prop != 0);
		prop = proposal->prop;
		inbound_spi = get_uint32((u_int32_t *)(prop + 1));
		if (inbound_spi != 0)
			ikev2_delete_sa(child_sa, prop->proto_id, remote, local,
					inbound_spi);
	}
}

/*
 * delete all ipsec sa of child_sa
 */
void
ikev2_child_delete_ipsecsa(struct ikev2_child_sa *child_sa)
{
	ikev2_child_delete_inbound(child_sa);
	ikev2_child_delete_outbound(child_sa);
}

/*
 * initiate Informational exchange with DELETE payload
 * and set state to EXPIRED when the exchange finishes
 */
static void ikev2_child_delete_callback(enum request_callback action,
					struct ikev2_child_sa *child_sa,
					void *data);

void
ikev2_child_delete(struct ikev2_child_sa *child_sa)
{
	struct ikev2_sa *ike_sa;
	struct ikev2_payloads *payl;
	struct prop_pair *proposal;
	vchar_t *delete_esp = 0;
	vchar_t *delete_ah = 0;
	struct ikev2_child_sa *exch_child_sa;

	TRACE((PLOGLOC, "ikev2_child_delete(%p)\n", child_sa));

	if (child_sa->delete_sent) {
		TRACE((PLOGLOC, "already sent\n"));
		return;
	}

	ike_sa = child_sa->parent;

	payl = racoon_malloc(sizeof(struct ikev2_payloads));
	ikev2_payloads_init(payl);

	for (proposal = child_sa->my_proposal[1];
	     proposal;
	     proposal = proposal->next) {
		struct isakmp_pl_p *prop;

		assert(proposal->prop != 0);
		prop = proposal->prop;
		switch (prop->proto_id) {
		case IKEV2PROPOSAL_ESP:
			if (delete_esp)
				isakmp_log(0, 0, 0, 0,
					   PLOG_INTERR, PLOGLOC,
					   "unexpected\n");
			delete_esp =
				ikev2_delete_payload(IKEV2_DELETE_PROTO_ESP,
						     sizeof(u_int32_t), 1,
						     (u_int8_t *)(prop + 1));
			if (!delete_esp)
				goto fail_nomem;
			TRACE((PLOGLOC, "delete esp spi=0x%x\n",
			       get_uint32((u_int32_t *)(prop + 1))));
			break;
		case IKEV2PROPOSAL_AH:
			if (delete_ah)
				isakmp_log(0, 0, 0, 0,
					   PLOG_INTERR, PLOGLOC,
					   "unexpected\n");
			delete_ah = ikev2_delete_payload(IKEV2_DELETE_PROTO_AH,
							 sizeof(u_int32_t),
							 1,
							 (u_int8_t *)(prop +
								      1));
			if (!delete_ah)
				goto fail_nomem;
			TRACE((PLOGLOC, "delete ah spi=0x%x\n",
			       get_uint32((u_int32_t *)(prop + 1))));
			break;
		default:
			isakmp_log(0, 0, 0, 0,
				   PLOG_INTERR, PLOGLOC,
				   "unexpected protocol %d\n", prop->proto_id);
			break;
		}
	}

	if (delete_esp) {
		ikev2_payloads_push(payl, IKEV2_PAYLOAD_DELETE,
				    delete_esp, TRUE);
		delete_esp = 0;
	}
	if (delete_ah) {
		ikev2_payloads_push(payl, IKEV2_PAYLOAD_DELETE,
				    delete_ah, TRUE);
		delete_ah = 0;
	}

	child_sa->delete_sent = TRUE;

	exch_child_sa =
	    ikev2_request_initiator_start(ike_sa,
				          ikev2_child_delete_callback,
				          payl);
	exch_child_sa->deleting_child_id = child_sa->child_id;

	ikev2_child_delete_inbound(child_sa);

      done:
	if (delete_esp)
		vfree(delete_esp);
	if (delete_ah)
		vfree(delete_ah);
	return;

      fail_nomem:
	isakmp_log(0, 0, 0, 0,
		   PLOG_INTERR, PLOGLOC, "memory allocation failure\n");
	goto done;
}

static void
ikev2_child_delete_callback(enum request_callback action,
			    struct ikev2_child_sa *child_sa, void *data)
{
	struct ikev2_child_sa *deleting_child_sa;

	TRACE((PLOGLOC,
	       "ikev2_child_delete_callback(%d, %p, %p)\n", action, child_sa,
	       data));
	switch (action) {
	case REQUEST_CALLBACK_CONTINUE:
		ikev2_informational_initiator_transmit(child_sa->parent,
						       child_sa,
						       (struct ikev2_payloads *)data);
		break;
	case REQUEST_CALLBACK_TRANSMIT_ERROR:
		/* none here */
		break;
	case REQUEST_CALLBACK_RESPONSE:
		ikev2_info_init_delete_recv(child_sa, (vchar_t *)data);
		deleting_child_sa = ikev2_find_child_by_id(child_sa->parent,
							   child_sa->deleting_child_id);
		if (deleting_child_sa)
			ikev2_child_state_set(deleting_child_sa,
					      IKEV2_CHILD_STATE_EXPIRED);
		else
			TRACE((PLOGLOC, "failed finding child_sa to delete\n"));
		break;
	default:
		isakmp_log(child_sa->parent, 0, 0, 0,
			   PLOG_INTERR, PLOGLOC,
			   "unknown action code %d\n", (int)action);
		break;
	}
}

/*
 * compute KEYMAT
 *
 *	 2.17 Generating Keying Material for CHILD_SAs
 *
 *	      KEYMAT = prf+(SK_d, Ni | Nr)
 *
 *	      KEYMAT = prf+(SK_d, g^ir (new) | Ni | Nr )
 *
 * INPUT:
 *	sa:		sa->sk_d for SK_d
 *		 	sa->n_i, sa->n_r if n_i and n_r are not provided
 *	g_ir:		g^ir if CREATE_CHILD_SA request had KE payload which contains g^i
 *	required_len:	required length of KEYMAT
 *	n_i, n_r:	nonces if this is CREATE_CHILD_SA exchange, NULL if IKE_AUTH exchange
 *
 * OUTPUT:
 *	returns `vchar_t *` which contains KEYMAT octets
 *	returns 0 if memory allocation failed
 *	returns 0 if required_len is too large  (>255*prf_output_len)
 */
static vchar_t *
compute_keymat(struct ikev2_sa *sa,
	       vchar_t *g_ir, size_t required_len, vchar_t *n_i, vchar_t *n_r)
{
	vchar_t *nonces = 0;
	int inputlen;
	u_int8_t *p;
	vchar_t *keymat = 0;

	/*
	 * (draft-17)
	 KEYMAT = prf+(SK_d, Ni | Nr)
	 
	 Where Ni and Nr are the Nonces from the IKE_SA_INIT exchange if this
	 request is the first CHILD_SA created or the fresh Ni and Nr from the
	 CREATE_CHILD_SA exchange if this is a subsequent creation.
	 */
	if (!n_i || !n_r) {
		n_i = sa->n_i;
		n_r = sa->n_r;
	}

	/*
	 * For CREATE_CHILD_SA exchanges including an optional Diffie-Hellman
	 * exchange, the keying material is defined as:
	 * 
	 * KEYMAT = prf+(SK_d, g^ir (new) | Ni | Nr )
	 */
	inputlen = n_i->l + n_r->l;
	if (g_ir)
		inputlen += g_ir->l;
	nonces = vmalloc(inputlen);
	if (!nonces)
		goto fail;
	p = (u_int8_t *) nonces->v;
	if (g_ir)
		VCONCAT(nonces, p, g_ir);
	VCONCAT(nonces, p, n_i);
	VCONCAT(nonces, p, n_r);

	keymat = ikev2_prf_plus(sa, sa->sk_d, nonces, required_len);

	/* done: */
      fail:
	if (nonces)
		vfree(nonces);
	return keymat;
}
